! den2cube:
!
! This program extracts information from a CASTEP formatted charge density
! file (.den_fmt) and the relevant .castep file to write the charge density as a .cube file.
!
! It is based on the castep2cube program developed by Jörg Meyer and
! the perl script den2grd developed by Keith Refson.
! Information about the .cube file format gathered from: http://paulbourke.net/dataformats/cube/
!
! Initial version: 12/08/2011.  Author: Aaron Hopkinson (a.hopkinson89@gmail.com)
!
! Summary of program:
!
! Subroutines:
!    filechecking :     Checks if the necessary files exist and also checks for an existing .cube file.
!    errorchecking:     Checks for errors in i/o and array (de)allocation and controls program accordingly.
!    castepsearch:      Searches the .castep file for the total number of ions, their atomic symbol and fractional positions.
!    castepsearcherror: Required by castepsearch to exit from a search. It checks for truncated data sets and finds the latest
!                       completed data set.
!    densearch:         Gets the real space lattice vectors, number of grid points and voxel data from a .den_fmt file.


program den2cube

  implicit none
  save

  character(len=80) :: filename                        ! Input/output filenames
  integer :: ios = 0                                   ! Tracks errors in i/o and allocation of arrays
  integer :: errortype                                 ! Decides which type of error we've encountered
  integer, dimension(3) :: gridpoints                  ! Grid points along a, b, c (na, nb, nc)
  double precision, dimension(1:3, 1:3) :: reallattice ! Real space lattice vectors
  integer :: novoxels                                  ! Number of voxels (na*nb*nc)
  integer :: noions = 0                                ! Total number of ions (Initialised to 0)
  character (len=2), allocatable, dimension(:) :: symb ! Name of ion from .castep
  double precision, allocatable, dimension(:,:) :: fracpos      ! Fractional positions of ion centres
  double precision, allocatable, dimension(:,:) :: tempfracpos  ! Temporary fracpos, used to find last set
  double precision, allocatable, dimension(:,:,:)  :: voxel     ! Stores the charges

  integer :: i, j                                      ! Loop counters
  integer :: x, y, z                                   ! Used for indexing voxel array

  double precision, parameter :: bohr2angstrom = 0.529177249 ! Unit conversion factor

  integer, allocatable, dimension(:) :: atomno         ! Atomic number (index for arrays)
  double precision, dimension(3) :: origin = 0.0       ! Origin of system
  double precision, allocatable, dimension(:,:) :: pos ! Absolute position of centre of ions

  ! The arrays below only go up to element 109. This should be enough.
  ! Note: If we need any of the len=3 ones, remember to
  ! change array length to 3 and add spaces to names below.
  ! Some formatting options may need changing as well.
  character(len=2), dimension (1:109) :: elements       = & ! Contains element symbols
       &(/ "H ", "He", "Li", "Be", "B ", "C ", "N ", "O ", &
       & "F ", "Ne", "Na", "Mg", "Al", "Si", "P ", "S ", &
       & "Cl", "Ar", "K ", "Ca", "Sc", "Ti", "V ", "Cr", &
       & "Mn", "Fe", "Co", "Ni", "Cu", "Zn", "Ga", "Ge", &
       & "As", "Se", "Br", "Kr", "Rb", "Sr", "Y ", "Zr", &
       & "Nb", "Mo", "Tc", "Ru", "Rh", "Pd", "Ag", "Cd", &
       & "In", "Sn", "Sb", "Te", "I ", "Xe", "Cs", "Ba", &
       & "La", "Ce", "Pr", "Nd", "Pm", "Sm", "Eu", "Gd", &
       & "Tb", "Dy", "Ho", "Er", "Tm", "Yb", "Lu", "Hf", &
       & "Ta", "W ", "Re", "Os", "Ir", "Pt", "Au", "Hg", &
       & "Tl", "Pb", "Bi", "Po", "At", "Rn", "Fr", "Ra", &
       & "Ac", "Th", "Pa", "U ", "Np", "Pu", "Am", "Cm", &
       & "Bk", "Cf", "Es", "Fm", "Md", "No", "Lr", "Rf", &
       & "Db", "Sg", "Bh", "Hs", "Mt" /)

  integer, dimension (1:109) :: valelec = &                 ! Number of valence electrons
       & (/ 1, 2, 1, 2, 3, 4, 5, 6, 7, 8, 1, 2, 3, 4, 5, 6, 7, 8, 1, & 
       & 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 2, 3, 4, 5, 6, 7, 8, 1, 2, & 
       & 3, 4, 5, 6, 7, 8, 9, 10, 1, 2, 3, 4, 5, 6, 7, 8, 1, 2, 3, & 
       & 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 3, 4, 5, 6, & 
       & 7, 8, 9, 10, 11, 2, 3, 4, 5, 6, 7, 8, 1, 2, 3, 4, 5, 6, 7, &
       & 8, 9, 10, 11, 12, 13, 14, 15, 16, 3, 4, 5, 6, 7, 8, 9 /)

  ! Get first argument and store as 'filename'.
  call getarg(1, filename)

  ! Check to see if argument is empty:
  if (trim(filename) .eq. '') then
     ! Tell user how to use the program:
     print *, "den2cube:"
     print *, "This program extracts the required information from a .castep file and a .den_fmt file ", &
          & "located in the current directory to create a .cube file which is outputted to the current directory."
     print *, ""
     print *, "To run this program, pass an argument to the program which the seedname of the input files."
     print *, ""
     print *, "For example: If this program is named den2cube and there are files named 'quartz.castep' ", &
          & "and 'quartz.den_fmt' in the current directory, you may use ./den2cube quartz to run the program."
     print *, ""
     stop
  else
     ! Use argument to check the files exist.
     call filechecking(filename)
  end if

  ! Search through the castep file for the number of ions, thier atomic symbols and their fractional positions.
  call castepsearch(filename, noions, ios, errortype)

  ! Search through the den_fmt file for the number of gridpoints, lattice vectors and voxels.
  call densearch(filename, gridpoints, reallattice, novoxels, ios, errortype)

  ! Scale the lattice vectors appropriately.
  do i = 1, 3
     do j = 1, 3
        reallattice(i,j) = reallattice(i,j)/(dble(gridpoints(i))*bohr2angstrom)
     end do
  end do

  ! Open stream to write to .cube file. Checks for errors in opening.
  open(unit=100, file=trim(filename)//'.cube', status='replace', action='write', iostat=ios)
  errortype = 21                ! 21 is opening .cube
  call errorchecking(errortype, ios, filename)

  ! Start writing the .cube file.
  errortype = 22                ! 22 is writing .cube
  write(100,'(a)',iostat=ios) trim(filename)
  call errorchecking(errortype, ios, filename)
  write(100,'(a)',iostat=ios) "Extracted from a CASTEP density formatted file using den2cube"
  call errorchecking(errortype, ios, filename)
  write(100,'(t3,i3,tr3,f9.6,tr3,f9.6,tr3,f9.6)',iostat=ios) noions,origin(1),origin(2),origin(3)
  call errorchecking(errortype, ios, filename)
  write(100,'(t3,i3,tr3,f9.6,tr3,f9.6,tr3,f9.6)',iostat=ios) gridpoints(1)+1,reallattice(1,1),reallattice(1,2),reallattice(1,3)
  call errorchecking(errortype, ios, filename)
  write(100,'(t3,i3,tr3,f9.6,tr3,f9.6,tr3,f9.6)',iostat=ios) gridpoints(2)+1,reallattice(2,1),reallattice(2,2),reallattice(2,3)
  call errorchecking(errortype, ios, filename)
  write(100,'(t3,i3,tr3,f9.6,tr3,f9.6,tr3,f9.6)',iostat=ios) gridpoints(3)+1,reallattice(3,1),reallattice(3,2),reallattice(3,3)
  call errorchecking(errortype, ios, filename)

  ! Allocate the array to the correct size, check for errors
  allocate(atomno(1:noions), pos(1:noions, 1:3), stat=ios)
  errortype = 31                ! 31 is array allocation
  call errorchecking(errortype, ios, filename)
  errortype = 22                ! 22 is writing .cube

  ! Loop over all ions.
  do i = 1, noions
     ! Loop over components.
     do j = 1,3
        ! Shift back to unit cell. Method taken from castep2cube (Thanks Jörg).
        if (fracpos(i,j) .ge. 0) then
           fracpos(i,j) = fracpos(i,j) - int(fracpos(i,j))
        else
           fracpos(i,j) = fracpos(i,j) -  int(fracpos(i,j)) + 1
        end if

        ! Scale results.
        pos(i, j) = fracpos(i,1)  * dble(gridpoints(1)) * reallattice(1,j) &
             & + fracpos(i,2)  * dble(gridpoints(2)) * reallattice(2,j)    &
             & + fracpos(i,3)  * dble(gridpoints(3)) * reallattice(3,j)

     end do

     ! Find the atomic number of each element (array lookup).
     ! Remember that if more elements are added we'll need to change this upper bound.
     do j = 1, 109
        if (symb(i) .eq. elements(j)) then
           atomno(i) = j
        end if
     end do

     ! From: http://paulbourke.net/dataformats/cube/
     ! The second number is 'unknown', it looks to be the number of valence electrons.
     ! Look up 'unknown' = no. valence electrons from array.
     write(100, '(t3, i3, tr3, f9.6, tr3, f9.6, tr3, f9.6, tr3, f9.6)', iostat=ios) &
          & atomno(i), dble(valelec(atomno(i))), pos(i,1), pos(i,2), pos(i,3)
     call errorchecking(errortype, ios, filename)

  end do

  ! We no longer need symb, fracpos, atomno, pos, reallattice. Deallocate arrays.

  deallocate (symb, fracpos, atomno, pos, stat=ios)
  errortype = 32                ! 32 is array deallocation
  call errorchecking(errortype, ios, filename)

  errortype = 22                ! 22 is writing .cube

  ! Scale the voxels.
  voxel = voxel/dble(novoxels)

  do x = 1, gridpoints(1)+1
     do y = 1, gridpoints(2)+1
        do z = 1, gridpoints(3)+1
           ! Copy edge voxels while we're here.
           voxel(gridpoints(1)+1,y,z) = voxel(1,y,z)
           voxel(x,gridpoints(2)+1,z) = voxel(x,1,z)
           voxel(gridpoints(1)+1,gridpoints(2)+1,z) = voxel(1,1,z)

           voxel(x,y,gridpoints(3)+1) = voxel(x,y,1)
           voxel(gridpoints(1)+1,y,gridpoints(3)+1) = voxel(1,y,1)
           voxel(x,gridpoints(2)+1,gridpoints(3)+1) = voxel(x,1,1)
        end do

        ! Thanks to Phil Hasnip for this next part:
        do z = 1, gridpoints(3)+1, 6
           write(100, '(t3, 6(e11.5,tr2))', iostat=ios) (voxel(x,y,z+i-1),i=1,min(6,gridpoints(3)+2-z))
           call errorchecking(errortype, ios, filename)
        end do
     end do
  end do

  close(100)

  ! Tell the user we're finished.
  print *, "File ", trim(filename), ".cube written successfully."

  ! Finally, let's free up some memory.
  deallocate (voxel, stat=ios)
  errortype = 32                ! 32 is array deallocation
  call errorchecking(errortype, ios, filename)

contains
  subroutine filechecking(filename)

    ! This subroutine checks if the necessary files exist and also checks for an existing .cube file.
    ! If an existing .cube file exists, it asks the user if they wish to overwrite it.

    implicit none

    character(len=80), intent(in) :: filename  ! Seedname of the files we're reading.
    character(len=1) :: overwrite = "x"        ! Do we want to overwrite .cube file if it already exists?
    logical :: exists = .true.                 ! Do the files exist? (Initially we'd hope so)


    ! Check to see if files to be read exist (.castep & .den_fmt).
    ! First, check for .castep:
    inquire(file=trim(filename)//'.castep', exist=exists)
    if (exists .eqv. .true.) then
       ! File exists, now look for density file:
       inquire(file=trim(filename)//'.den_fmt', exist=exists)
       if (exists .eqv. .true.) then
          ! Both files exist, now let's check to see if the .cube file already exists:
          inquire(file=trim(filename)//'.cube', exist=exists)
          if (exists .eqv. .true.) then
             do while (overwrite .eq. 'x')
                ! Ask user if they want to overwrite file:
                print *, "The file: ", trim(filename), ".cube already exists."
                print *, "Do you wish to overwrite this file? (y/n)"
                read *, overwrite
                if (overwrite .eq. 'n') then
                   print *, "Nothing has been written. Please rename", &
                        & " or remove the ", trim(filename), &
                        & ".cube file and restart the ", &
                        & "program to continue."
                   stop
                else if (overwrite .eq. 'y') then
                   print *, "This program will overwrite the ", &
                        & trim(filename), ".cube file."
                   ! Continue in main program from here.
                else
                   ! If either y/n - user cannot follow instructions. Loop until they get it right.
                   print *, "Please enter either y or n."
                   overwrite = "x"
                end if
             end do
          end if
          ! If .cube file doesn't exist then we simply create a new one.
          ! Nothing else is needed here.
       else
          ! The .castep file exists, the density file doesn't.
          print *, "The file: ", trim(filename), ".den_fmt does not exist."
          print *, "Please restart this program from a directory containing this file ", &
               & "and a ", trim(filename), ".castep file."
          stop
       end if
    else
       ! The .castep file doesn't exist. Don't even bother looking for the density file.
       print *, "The file: ", trim(filename), ".castep does not exist."
       print *, "We did not look for a .den_fmt file."
       print *, "Please restart this program from a directory containing this file ", &
            & "and a ", trim(filename), ".den_fmt file."
       stop
    end if
  end subroutine filechecking


  subroutine errorchecking(errortype, ios, filename)

    ! This subroutine checks for errors in i/o and array (de)allocation.
    ! The value of errortype defines which type of errors are being checked and therefore which action to take.
    !
    ! errortype values:
    ! 1  - opening of .castep file
    ! 11 - opening of .den_fmt file
    ! 12 - reading of .den_fmt file
    ! 21 - opening of .cube file
    ! 22 - writing of .cube file
    ! 31 - allocation of arrays
    ! 32 - deallocation of arrays

    implicit none

    integer, intent(in) :: ios                   ! Error code .ne. 0 is an error
    integer, intent(in) :: errortype             ! Type of error we're checking for
    character(len=80), intent(in) :: filename    ! Seedname of the files we're reading.

    if (ios .ne. 0) then
       select case (errortype)
       case (1)
          print *, "Error opening: ", trim(filename), ".castep"
          print *, "Please ensure the file is not corrupt and run this program again."
          stop
       case (11)
          print *, "Error opening: ", trim(filename), ".den_fmt"
          print *, "Please ensure the file is not corrupt and run this program again."
          stop
       case (12)
          print *, "Error reading from: ", trim(filename), ".den_fmt"
          stop
       case (21)
          print *, "Error opening: ", trim(filename), ".cube"
          print *, "Please ensure there is sufficient free space to write this file."
          stop
       case (22)
          print *, "Error writing to: ", trim(filename), ".cube"
          print *, "Please ensure there is sufficient free space to write this file."
          stop
       case (31)
          print *, "Error allocating arrays. There is not enough free space to allocate arrays."
          print *, "Please free up some memory and run this program again."
          stop
       case (32)
          print *, "Error deallocating arrays."
       end select
    end if

  end subroutine errorchecking

  subroutine castepsearch(filename, noions, ios, errortype)

    ! This subroutine searches the .castep file for the total number of ions,
    ! their atomic symbol and their fractional positions.

    implicit none

    character(len=80), intent(in) :: filename   ! Which file to extract data from
    integer, intent(out) :: noions              ! Number of ions
    integer, intent(out) :: ios                 ! i/o errors
    integer, intent(out) :: errortype           ! Type of error to check for

    logical :: copy = .true.                    ! Whether to copy from the tempfracpos -> fracpos
    logical :: run = .true.                     ! Whether to search .castep file
    character(len=80) :: line                   ! Current line of the .castep file
    integer :: search = 0                       ! Results of search (index function)
    integer :: i, j, k                          ! Loop counters


    ! Open stream to .castep file, check for errors.
    open(unit=10, file=trim(filename)//'.castep', status='old', action='read', iostat=ios)
    errortype = 1                ! 1 is opening of .castep file
    call errorchecking(errortype, ios, filename)

    ! Initially search through file.
    ! Eventually this routine will give a non-zero error code when the end of the file is reached.
    ! This calls the errorchecking subroutine which causes run to go to false, breaking the loop and using the most
    ! recent set of data (fractional positions).
    do while (run .eqv. .true.)
       read(10, '(a)', iostat=ios) line   ! Read line and store
       call castepsearcherror(ios, noions, run, copy)

       ! Have we found the number of ions yet?
       if (noions .eq. 0) then
          ! No. Define search to find them
          search = index(line, 'Total number of ions in cell')

          ! If we find the number, save it
          if (search .eq. 26) then
             read(line, '(t56, i10)', iostat=ios) noions !mijp: this line written by cell with fmt=15x,a40,i5
             call castepsearcherror(ios, noions, run, copy)
             ! Now we know number of ions, allocate arrays
             allocate (symb(1:noions), fracpos(1:noions, 1:3), tempfracpos(1:noions, 1:3), stat=ios)
             errortype = 31
             call errorchecking(errortype, ios, filename)

             ! Initialise fractional position arrays to 0
             fracpos = 0.0
             tempfracpos = 0.0
          end if
       else
          ! Yes, we've found the number of ions, now look for rest of data
          ! (ie: names of atoms and their position)
          ! Start by defining search
          search = index(line, '            x  ')
          if (search .eq. 1) then
             ! We've found the beginning of the rest of the data. Now skip 2 lines
             read (10, *, iostat=ios)
             call castepsearcherror(ios, noions, run, copy)
             read (10, *, iostat=ios)
             call castepsearcherror(ios, noions, run, copy)
             ! Data starts here. Loop over all ions to store info
             do i = 1, noions
                ! Initially store the fractional positions in a temporary array.
                ! (Remember, if we use a 3 character array for elements we need to change the
                ! next statement to 'a3').
                !mijp: this line was written by cell with fmt=(11x,' x  ',a8,i6,6x,3f11.6,'   x')
                read(10, '(t16, a2, t36, 3f11.6)', iostat=ios) symb(i), tempfracpos(i,1), &
                     & tempfracpos(i,2), tempfracpos(i,3)

                ! When we get to the last ion, run error checking to see if there were errors reading the
                ! fractional positions
                if (i .eq. noions) then
                   call castepsearcherror(ios, noions, run, copy)
                   ! By default, copy is true and the temporary fractional positions are copied into the final
                   ! fractional position array at the end of each run.
                   ! If error checking gives an error and data cannot be extracted,
                   ! the program will end.  If we encounter more than one set of positions, the most recent is
                   ! used. But, if we happen to find a truncated data set, we revert back to the last full set.
                   ! Reverting back to the last set occurs when copy is false.
                   if (copy .eqv. .true.) then
                      ! Copy from the temporary array into the final array.
                      do j = 1, noions
                         do k = 1, 3
                            fracpos(j,k) = tempfracpos(j,k)
                         end do
                      end do
                   end if
                   ! Reset the temporary array for the next loop.
                   tempfracpos = 0.0
                   ! fracpos now contains the most recent completed set of fractional positions.
                end if
             end do
          end if
       end if
    end do
    close(10)
  end subroutine castepsearch

  subroutine castepsearcherror(ios, noions, run, copy)

    ! This subroutine checks for trucation of the .castep file.  If there is no data found
    ! by the end of the file, the program fails. If the file terminates part-way through a data
    ! set and there is already a previously completed set in memory then the program falls back
    ! and uses this set.

    implicit none

    integer, intent(in) :: ios                                               ! i/o errors
    integer, intent(in) :: noions                                            ! Total number of ions
    logical, intent(out) :: run                                              ! For searching the .castep file
    logical, intent(out) :: copy                                             ! Should we copy the temporary fracpos to the final fracpos?

    integer :: i, j                                                          ! Loop counters
    logical :: fracposempty, tempfracposempty                                ! Are the fractional position arrays empty?

    if (ios .ne. 0) then
       ! If we have not found any ions, then there is no data in the .castep file.
       if (noions .eq. 0) then
          print *, "The required data could not be read from the .castep file."
          print *, "Please check that the file has not been truncated or corrupt and restart the program."
          stop
       else
          ! We have found the number of ions, now check to see if there is anything stored in either
          ! if the two fractional position arrays.
          tempfracposempty = .true. !mijp - must assign an initial value
          fracposempty = .true.     !mijp - must assign an initial value
          do i = 1, noions
             do j = 1, 3
                ! Are any elements in tempfracpos non-zero?
                if (tempfracpos(i,j) .ne. 0.0) then
                   tempfracposempty = .false.
                end if
                ! Are any elements in fracpos non-zero?
                if (fracpos(i,j) .ne. 0.0) then
                   fracposempty = .false.
                end if

                ! At the end of the run:
                if ((i .eq. noions) .and. (j .eq. 3)) then
                   ! Does fracpos store anything?
                   if (fracposempty) then
                      print *, "The fractional positions could not be found in the .castep file."
                      print *, "Please check that the file has not been truncated or corrupt and restart the program."
                      stop
                   else
                      ! If tempfracpos is empty, then fracpos should now be storing a complete set.
                      if (tempfracposempty) then
                         ! This is good, now we can stop looking through the .castep file.
                         run = .false.
                      else
                         ! tempfracpos is not empty, we must have hit a read error before the next
                         ! fractional position set was completed, we can revert back to the original set.
                         print *, "It appears your data was truncated part way through a run."
                         print *, "A set of fractional positions has already been found. We shall revert to this set."
                         ! We don't want to copy the incomplete set but we do want to stop searching the .castep file.
                         copy = .false.
                         run = .false.
                      end if
                   end if
                end if
             end do
          end do
       end if
    end if
  end subroutine castepsearcherror

  subroutine densearch(filename, gridpoints, reallattice, novoxels, ios, errortype)

    ! This subroutine gets the real space lattice vectors, number of grid points and voxel data from a .den_fmt file.

    implicit none

    character(len=80), intent(in) :: filename                              ! File to search through
    integer, dimension(3), intent(out) :: gridpoints                       ! Number of grid points along x,y,z
    double precision, dimension(1:3, 1:3), intent(out) :: reallattice      ! Real space lattice vectors
    integer, intent(out) :: novoxels                                       ! Number of voxels (na*nb*nc)

    integer, intent(out) :: ios                                            ! i/o error codes
    integer, intent(out) :: errortype                                      ! Type of error

    integer :: i, j                                                        ! Loop counters
    integer :: x, y, z                                                     ! Voxel indices


    ! Extract from .den file, first open stream - check for errors
    open(unit=20, file=trim(filename)//'.den_fmt', status='old', action='read', iostat=ios)
    errortype = 11              ! 11 is for opening den_fmt
    call errorchecking(errortype, ios, filename)
    errortype = 12              ! 12 is for reading den_fmt

    ! Loop over the file
    do i = 1,6
       read(20, *, iostat=ios)
       call errorchecking(errortype, ios, filename)
       ! First 3 lines are useless to us.
       ! On the 3rd run, extract the lattice vectors.
       if (i .eq. 3) then
          read(20, *, iostat=ios) reallattice(1,1), reallattice(1,2), reallattice(1,3)
          call errorchecking(errortype, ios, filename)
          read(20, *, iostat=ios) reallattice(2,1), reallattice(2,2), reallattice(2,3)
          call errorchecking(errortype, ios, filename)
          read(20, *, iostat=ios) reallattice(3,1), reallattice(3,2), reallattice(3,3)
          call errorchecking(errortype, ios, filename)
       end if

       ! 7th and 8th lines useless to us.
       ! On 5th search, store the number of gridpoints along x,y,z
       if (i .eq. 5) then
          read(20, *, iostat=ios) gridpoints(1), gridpoints(2), gridpoints(3)
          call errorchecking(errortype, ios, filename)
          ! Now allocate the voxel array to the right size.
          errortype = 31                ! 31 is allocation of arrays
          allocate(voxel(1:gridpoints(1)+1, 1:gridpoints(2)+1, 1:gridpoints(3)+1), stat=ios)
          call errorchecking(errortype, ios, filename)
          errortype = 12                ! 12 is reading den_fmt
       end if

       ! 10th and 11th lines useless.
       if (i .eq. 6) then
          read(20, *, iostat=ios)
          call errorchecking(errortype, ios, filename)
          ! Work out number of loops to do:
          novoxels = gridpoints(1)*gridpoints(2)*gridpoints(3)
          ! Loop over number of lines to find voxels:
          do j = 1, novoxels
             ! Store in the right place of the voxel array.
             read(20, *, iostat=ios) x, y, z, voxel(x,y,z)
             call errorchecking(errortype, ios, filename)
          end do
       end if
    end do
    close(20)
  end subroutine densearch

end program den2cube
